/*
 *
 *  Copyright (C) 1999-2001 Silicon Graphics, Inc.
 *
 *  Written by John Hawkes (hawkes@sgi.com)
 *  Based on lockstat.c by Jack Steiner (steiner@sgi.com)
 *
 *  Modifications by Ray Bryant (raybry@us.ibm.com) Feb-Mar 2000
 *  Changes Copyright (C) 2000 IBM, Inc.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <linux/utsname.h>

#include "lockstat.h"
#include "getsetdata.h"

static int dataFile = 0;
static int isProcfileFlag = 0;
static char *dataFilename;

static int mirrorOutFile = 0;
static char *mirrorOutFilename = NULL;

static char *generalReadBuf = NULL;
static int  generalReadBufLen = 0;

static int currentState = LSTAT_OFF;
static int maxCpusCount = 0;

static int header_size;
static int max_stat_index = 0;
static int max_read_lock_index = 0;

/*
 * must match what is in include/linux/lockmeter.h
 */
char *command_name[] = {"OFF","ON","RESET","RELEASED",""};

void
closeFiles()
{
    int result;

    /*
     *  The only file that needs to be closed is the mirrored output file,
     *  if it exists.
     */
    if (mirrorOutFile) {
	close(mirrorOutFile);

	/* fix up file permissions */
	result = chmod(mirrorOutFilename,S_IWUSR | S_IRUSR | S_IRGRP | S_IROTH);
	if (result) {
	    fprintf(stderr,"%s: cannot chmod '%s'\n", __func__, mirrorOutFilename);
	}
    }
}

/*
 * getPerCpuLockCounts:
 *
 *	Extract the per cpu lock counts from the general read
 *	buffer.
 */
size_t
getPerCpuLockCounts(
	void *memcpySrcAddr,
	lstat_lock_counts_t ***		cpu_counts_ptr)
{
    lstat_lock_counts_t      **perCpuLockCounts;
    size_t  memcpyLen;

    int cpu;

    if (*cpu_counts_ptr == NULL) {
	perCpuLockCounts = (lstat_lock_counts_t **)
		calloc(maxCpusCount, sizeof(lstat_lock_counts_t*));
	if (perCpuLockCounts == NULL) {
	    die(1,"%s: cannot malloc %d bytes for"
		    " lstat_lock_counts_t per cpu pointer array",
		    __func__, maxCpusCount * sizeof(lstat_lock_counts_t*));
	}
	*cpu_counts_ptr = perCpuLockCounts;
    } else
	perCpuLockCounts = *cpu_counts_ptr;

    /*
     * extract each cpu's lock counts
     */
    memcpyLen = max_stat_index * sizeof(lstat_lock_counts_t);
    for(cpu = 0; cpu < maxCpusCount; ++cpu) {
	lstat_lock_counts_t *counts = perCpuLockCounts[cpu];
	if (!counts) {
	    counts = (lstat_lock_counts_t *)malloc(memcpyLen);
	    if (!counts) {
		die(1,"%s: cannot malloc %d bytes for"
		    " cpu %d lstat_lock_counts_t buffer",
		    __func__, cpu, memcpyLen);
	    }
	    perCpuLockCounts[cpu] = counts;
	}
        memcpy(counts, memcpySrcAddr, memcpyLen);
	memcpySrcAddr += memcpyLen;
    }
    return (maxCpusCount * memcpyLen);
}

/*
 * getPerCpuReadLockCounts:
 *
 *	Extract the per cpu read lock counts from the general read
 *	buffer.
 */
size_t
getPerCpuReadLockCounts(
	void *memcpySrcAddr,
	lstat_read_lock_counts_t ***		cpu_counts_ptr)
{
    lstat_read_lock_counts_t      **perCpuLockCounts;
    size_t  memcpyLen;

    int cpu;

    if (*cpu_counts_ptr == NULL) {
	perCpuLockCounts = (lstat_read_lock_counts_t **)
		calloc(maxCpusCount, sizeof(lstat_read_lock_counts_t*));
	if (perCpuLockCounts == NULL) {
	    die(1, "%s: cannot malloc %d bytes for"
		    " lstat_read_lock_counts_t per cpu pointer array",
		    __func__, maxCpusCount * sizeof(lstat_read_lock_counts_t*));
	}
	*cpu_counts_ptr = perCpuLockCounts;
    } else
	perCpuLockCounts = *cpu_counts_ptr;

    /*
     * extract each cpu's lock counts
     */
    memcpyLen = max_read_lock_index * sizeof(lstat_read_lock_counts_t);
    for(cpu = 0; cpu < maxCpusCount; ++cpu) {
	lstat_read_lock_counts_t *counts = perCpuLockCounts[cpu];
	if (!counts) {
	    counts = (lstat_read_lock_counts_t *)malloc(memcpyLen);
	    if (!counts) {
		die(1, "%s: cannot malloc %d bytes for"
		    " cpu %d lstat_read_lock_counts_t buffer",
		    __func__, cpu, memcpyLen);
	    }
	    perCpuLockCounts[cpu] = counts;
	}
        memcpy(counts, memcpySrcAddr, memcpyLen);
	memcpySrcAddr += memcpyLen;
    }
    return (maxCpusCount * memcpyLen);
}


/*
 *  getKernelData
 *
 *	This is the general interface to read lockmetering data from a
 *	lockmeter-capable kernel.
 *	The args are pointers to pointers of buffers.  If the caller
 *	supplies a non-null pointer to a buffer, then the appropriate data
 *	is passed back to the caller; if the caller supplies no buffers,
 *	then getKernelData() malloc's buffers and returns those back to
 *	the user.  It is the caller's responsibility to reuse these buffers
 *	on a subsequent call, or to free() the buffers.
 *	If the caller specifies readHeaderOnly=1, then only the initial
 *	lstat_user_request_t information is returned; otherwise, all four
 *	subparts of the data is returned.
 */
void
getKernelData(
	lstat_user_request_t **		request_ptr,
	lstat_lock_counts_t ***		cpu_counts_ptr,
	lstat_directory_entry_t **	directory_ptr,
	lstat_read_lock_counts_t ***	read_lock_counts_ptr,
	int				readHeaderOnly)
{
    void *memcpySrcAddr;
    lstat_read_lock_counts_t **perCpuReadLockCounts;

    size_t	read_count;
    size_t buf_size;
    size_t  memcpyLen;

    if (!dataFile) {
	die(1, "%s: data file is not open!\n", __func__);
    }

    if (readHeaderOnly) {
	/*
	 * Read "header" [lstat_user_request_t] to size data buffer
	 */
	buf_size = sizeof(lstat_user_request_t);  /* easier to read */
	if (*request_ptr == NULL) {
	    *request_ptr = (lstat_user_request_t *)malloc(buf_size);
	    if (*request_ptr == NULL) {
		die(1, "%s: cannot malloc %d bytes for"
			" lstat_user_request_t buffer", __func__, buf_size);
	    }
	}
	read_count = read(dataFile, *request_ptr, buf_size);
	if (read_count != buf_size) {
	    die(1, "%s: "
		"read of %d bytes of lstat_user_request_t from file '%s' "
		"gets %d bytes\n", __func__,
		buf_size, dataFilename, read_count);
	}

	/*
	 *  If reading from something other than the /proc file,
	 *  then rewind to undo the previous read()
	 */
	if (!isProcfileFlag) {
	    read_count = (int) lseek(dataFile, -buf_size, SEEK_CUR);
	    if (read_count < 0) {
		die(1, "%s: lseek failure\n", __func__);
	    }
	}

	currentState = (*request_ptr)->state;
	header_size  = (*request_ptr)->header_size;
	max_stat_index = (*request_ptr)->max_stat_index;
	max_read_lock_index = (*request_ptr)->max_read_lock_index;

	/*
	 * For now, complain if # of cpus changes.
	 * TODO:  handle add/remove of cpu during lockmeter
	 */
	if (maxCpusCount == 0)
	    maxCpusCount = (*request_ptr)->maxcpus;
	else if (maxCpusCount != (*request_ptr)->maxcpus) {
	    die(1, "getCollectionState: "
		"supposedly constant 'maxcpus' has changed from %d to %d!\n",
		maxCpusCount, (*request_ptr)->maxcpus);
	}

	/*
	 * If we don't have a big general read buffer yet, alloc it now.
	 * Use actual header_size from initial read of the header itself
	 * in case of size mismatch [e.g., utsname funkiness].
	 */
	if (!generalReadBuf) {
	    generalReadBufLen = header_size  +
	      + (maxCpusCount * max_stat_index * sizeof(lstat_lock_counts_t))
	      + (max_stat_index * sizeof(lstat_directory_entry_t))
	      + (maxCpusCount * max_read_lock_index
				      * sizeof(lstat_read_lock_counts_t));

	    generalReadBuf = (char *)malloc(generalReadBufLen);
	    if (!generalReadBuf) {
		die(1, "%s:"
		    "Unable to malloc %d bytes for generalReadBuf\n",
		    __func__, generalReadBufLen);
	    }
	}

	return;	/* done! */

    } /* readHeaderOnly */

    if (!generalReadBuf) {
	die(1,"%s: first use must be readHeaderOnly\n", __func__);
    }

    /*
     * Read the full concatenated data from the kernel
     */
    read_count = read(dataFile, generalReadBuf, generalReadBufLen);
    if (read_count != generalReadBufLen) {
	die(1, "%s: "
		"read requested %d bytes, received %d\n",
		__func__, generalReadBufLen, read_count);
    }

    /*
     * Mirror the input data, if requested
     */
    if (mirrorOutFilename) {
	int write_count;
	if (!mirrorOutFile) {
	    mirrorOutFile = open(mirrorOutFilename, O_RDWR|O_CREAT);
	    if (mirrorOutFile < 0) {
		die(1, "%s: "
			"cannot open mirror output file '%s', errno=%d\n",
			__func__, mirrorOutFilename, errno);
	    }
	}
	write_count = write(mirrorOutFile, generalReadBuf, generalReadBufLen);
	if (write_count != read_count) {
	    die(1, "%s: attempted to write %d bytes "
		    "to mirror file, only %d bytes succeeded\n",
		    __func__, generalReadBufLen, write_count);
	}
    }

    /*
     *  Separate the data into the various discrete chunks.
     *  Malloc buffers as necessary.
     */

    /*
     * The "header" or "User Request" structure
     */
    memcpySrcAddr = generalReadBuf;
    memcpyLen = header_size;	/* actual size */
    if (*request_ptr == NULL) {
	*request_ptr = (lstat_user_request_t *)malloc(memcpyLen);
	if (*request_ptr == NULL) {
	    die(1, "%s: cannot malloc %d bytes for"
		    " lstat_user_request_t buffer", __func__, memcpyLen);
	}
    }
    memcpy((void *)*request_ptr, memcpySrcAddr, memcpyLen);
    memcpySrcAddr += memcpyLen;		/* bump for next memcpy */

    /*
     * The lock counts
     */
    memcpyLen = getPerCpuLockCounts(memcpySrcAddr, cpu_counts_ptr);
    memcpySrcAddr += memcpyLen;

    /*
     * The lockstat directory
     */
    memcpyLen = max_stat_index * sizeof(lstat_directory_entry_t);
    if (*directory_ptr == NULL) {
	*directory_ptr = (lstat_directory_entry_t *)malloc(memcpyLen);
	if (*directory_ptr == NULL) {
	    die(1, "%s: cannot malloc %d bytes for"
		    " lstat_directory_entry_t buffer", __func__, memcpyLen);
	}
    }
    memcpy((void *)*directory_ptr, memcpySrcAddr, memcpyLen);
    memcpySrcAddr += memcpyLen;

    /*
     * The read lock counts
     */
    memcpyLen = getPerCpuReadLockCounts(memcpySrcAddr, read_lock_counts_ptr);
}

int
isProcFile()
{
    return isProcfileFlag;
}

/*
 * openKernelData
 *	Return >=0 if kernel data file can be opened, -1 if it cannot.
 *	Open read-write if /proc/lockmeter node, readonly otherwise.
 */
int
openKernelData(char *filename)
{
    dataFilename = (filename) ? filename : defaultDataFilename;
    isProcfileFlag = (strcmp(dataFilename, defaultDataFilename) == 0);

    /* need to be root if reading /proc/lockmeter */
    if (isProcfileFlag)
	setuid(0);

    dataFile = open(dataFilename, (isProcfileFlag) ? O_RDWR : O_RDONLY);
    return dataFile;
}

void
setCollectionState(int newState)
{
    char writeBuf[1];

    switch(newState) {
	case LSTAT_ON:
		if (currentState == LSTAT_ON) {
			die(1, "Lockmeter statistics already ON.\n");
		}
		break;
	case LSTAT_OFF:
		if (currentState == LSTAT_OFF) {
			die(1, "Lockmeter statistics already OFF.\n");
		}
		break;
	case LSTAT_RESET:
		if (currentState == LSTAT_ON) {
			die(1, "Lockmeter statistics are ON;"
				" turn them off before reset.\n");
		}
		break;
	case LSTAT_RELEASE:
		if (currentState == LSTAT_ON) {
			die(1, "Lockmeter statistics ON;"
				" turn them off before release.\n");
		}
		break;
	default:
		die(1, "setCollectionState: unknown newState=%d\n",
			newState);
    }

    writeBuf[0] = newState;
    if (write(dataFile, writeBuf, 1) != 1) {
	die(1,"%s:"
		" error %d setting lockmeter statistics to:%d\n",
		__func__, errno, newState);
    } else {
	printf("Lockmeter statistics are now %s\n", command_name[newState]);

	if (newState == LSTAT_ON)
		currentState = LSTAT_ON;
	else if (newState == LSTAT_OFF)
		currentState = LSTAT_OFF;
    }
}

void
setMirrorOutFilename(char *filename)
{
    mirrorOutFilename = (char *)malloc(strlen(filename) + 1);
    if (!mirrorOutFilename) {
	die(1, "%s: cannot malloc %d bytes\n",
		strlen(filename)+1);
    }
    strcpy(mirrorOutFilename, filename);
}
