/*
 *  Copyright (C) 1999-2002 Silicon Graphics, Inc.
 *
 *  Written by John Hawkes (hawkes@sgi.com)
 *  Based on klstat.h by Jack Steiner (steiner@sgi.com)
 *
 *  Modified by Ray Bryant (raybry@us.ibm.com) Feb-Apr 2000
 *  Changes Copyright (C) 2000 IBM, Inc.
 *  Added save of index in spinlock_t to improve efficiency
 *  of "hold" time reporting for spinlocks
 *  Added support for hold time statistics for read and write
 *  locks.
 *  Moved machine dependent code to include/asm/lockmeter.h.
 *
 *  Modified by Lee Schermerhorn (lee.schermerhorn@hp.com)
 *  ~Fall 2006 to separate kernel internal part of lockmeter.h
 *  from user<=>kernel API.
 */

#ifndef _LINUX_LOCKMETER_H
#define _LINUX_LOCKMETER_H

#ifdef CONFIG_LOCKMETER

/*
 * architecture-independent user<=>kernel lockmeter API
 */

/*
 * raybry -- version 2: added efficient hold time statistics
 *           requires lstat recompile, so flagged as new version
 * raybry -- version 3: added global reader lock data
 * hawkes -- version 4: removed some unnecessary fields to simplify mips64 port
 * version 5 ???
 * lts    -- version 6: add sizing info for dynamic table sizing
 */
#define LSTAT_VERSION	6

/*
 * Currently, the mips64 and sparc64 kernels talk to a 32-bit lockstat, so we
 * need to force compatibility in the inter-communication data structure.
 */

#if defined(CONFIG_MIPS32_COMPAT)
#define TIME_T		uint32_t
#elif defined(CONFIG_SPARC) || defined(CONFIG_SPARC64)
#define TIME_T		uint64_t
#else
#define TIME_T		time_t
#endif

#if defined(__KERNEL__) ||             \
   ( !defined(CONFIG_MIPS32_COMPAT) && \
     !defined(CONFIG_SPARC) &&         \
     !defined(CONFIG_SPARC64)     ) || \
   (_MIPS_SZLONG==32)
#define POINTER		void *
#else
#define	POINTER		int64_t
#endif

/*
 * User request to:
 *	- turn statistic collection on/off, or to reset
 */
#define LSTAT_OFF	 0
#define LSTAT_ON	 1
#define LSTAT_RESET      2
#define LSTAT_RELEASE    3

/*
 * lstat_utsname:  lockmeter private version of uts_name struct so
 * that neither lockmeter command does not need to include a uts_name
 * header.  Sizes lifted from <linux/utsname.h> struct new_utsname.
 */
#define LSTAT_UTSNAME_LEN 65
struct lstat_utsname {
	char sysname[LSTAT_UTSNAME_LEN];
	char nodename[LSTAT_UTSNAME_LEN];
	char release[LSTAT_UTSNAME_LEN];
	char version[LSTAT_UTSNAME_LEN];
	char machine[LSTAT_UTSNAME_LEN];
	char domainname[LSTAT_UTSNAME_LEN];
};

/*
 * lstat_user_request:  describes configuration and in-kernel state of
 * lockmeter.   Used by lockstat command to extract lock statistics
 * and generate reports.
 */
typedef struct lstat_user_request {
	short	lstat_version;		/* version of the data */
	short	state;			/* the current state is returned */
	int	header_size;		/* size of this header */

	int	maxcpus;		/* Number of cpus present */
	int	max_stat_index;         /* size of lock stat table */
	int	max_read_lock_index;    /* size of read lock stat table */

	int	next_free_dir_index;	/* index of the next free directory entry */
	TIME_T	first_started_time;	/* when measurement enabled for first time */
	TIME_T	started_time;		/* time in secs since 1969 when stats last turned on  */
	TIME_T	ending_time;		/* time in secs since 1969 when stats last turned off */
	uint32_t cycleval;		/* cycles per second */
#ifdef notyet
	void	*kernel_magic_addr;	/* address of kernel_magic */
	void	*kernel_end_addr;	/* contents of kernel magic (points to "end") */
#endif
	int	next_free_read_lock_index; /* index of next (global) read lock stats struct */
	uint64_t started_cycles64;	/* cycles when measurement last started        */
	uint64_t ending_cycles64;	/* cycles when stats last turned off           */
	uint64_t enabled_cycles64;	/* total cycles with measurement enabled       */
	int	intervals;		/* number of measurement intervals recorded      */
						/* i.e. number of times we did lstat on;lstat off*/
	int	dir_overflow;		/* number of times we wanted more space in directory */
	int	rwlock_overflow;	/* # of times we wanted more space in read_locks_count */
	struct lstat_utsname uts;		/* info about machine where stats are measured */
						/* -T option of lockstat allows data to be     */
						/* moved to another machine. ................. */
} lstat_user_request_t;

/*
 * Values for the "action" parameter passed to lstat_update.
 *	ZZZ - do we want a try-success status here???
 */
#define LSTAT_ACT_NO_WAIT	0
#define LSTAT_ACT_SPIN		1
#define LSTAT_ACT_REJECT	2
#define LSTAT_ACT_WW_SPIN       3
#define LSTAT_ACT_SLEPT		4 /* UNUSED */

#define LSTAT_ACT_MAX_VALUES	4 /* NOTE: Increase to 5 if use ACT_SLEPT */

/*
 * Special values for the low 2 bits of an RA passed to
 * lstat_update.
 *
 * We use these values to figure out what kind of lock data
 * is stored in the statistics table entry at index .......
 */
#define LSTAT_RA_SPIN           0  /* spin lock data */
#define LSTAT_RA_READ           1  /* read lock statistics */
#define LSTAT_RA_SEMA		2  /* RESERVED */
#define LSTAT_RA_WRITE          3  /* write lock statistics*/
#define LSTAT_RA_MASK           3UL

/*
 * Constants used for lock addresses in the lstat_directory
 * to indicate special values of the lock address.
 */
#define	LSTAT_MULTI_LOCK_ADDRESS	NULL


/*
 *	This defines an entry in the lockstat directory. It contains
 *	information about a lock being monitored.
 *	A directory entry only contains the lock identification -
 *	counts on usage of the lock are kept elsewhere in a per-cpu
 *	data structure to minimize cache line pinging.
 */
typedef struct lstat_directory_entry {
	POINTER	caller_ra;		  /* RA of code that set lock */
	POINTER	lock_ptr;		  /* lock address */
	ushort	next_stat_index;  /* Used to link multiple locks that have the same hash table value */
} lstat_directory_entry_t;

/*
 *	A multi-dimensioned array used to contain counts for lock accesses.
 *	The array is 3-dimensional:
 *		- CPU number. Keep from thrashing cache lines between CPUs
 *		- Directory entry index. Identifies the lock
 *		- Action. Indicates what kind of contention occurred on an
 *		  access to the lock.
 *
 *	The index of an entry in the directory is the same as the 2nd index
 *	of the entry in the counts array.
 */
/*
 *  This table contains data for spin_locks, write locks, and read locks
 *  Not all data is used for all cases.  In particular, the hold time
 *  information is not stored here for read locks since that is a global
 *  (e. g. cannot be separated out by return address) quantity.
 *  See the lstat_read_lock_counts_t structure for the global read lock
 *  hold time.
 */
typedef struct lstat_lock_counts {
	uint64_t    cum_wait_ticks;	/* sum of wait times               */
	                                /* for write locks, sum of time a  */
					/* writer is waiting for a reader  */
	int64_t	    cum_hold_ticks;	/* cumulative sum of holds         */
	                                /* not used for read mode locks    */
					/* must be signed. ............... */
	uint32_t    max_wait_ticks;	/* max waiting time                */
	uint32_t    max_hold_ticks;	/* max holding time                */
	uint64_t    cum_wait_ww_ticks;  /* sum times writer waits on writer*/
	uint32_t    max_wait_ww_ticks;  /* max wait time writer vs writer  */
	                                /* prev 2 only used for write locks*/
	uint32_t    acquire_time;       /* time lock acquired this CPU     */
	uint32_t    count[LSTAT_ACT_MAX_VALUES];
} lstat_lock_counts_t;

/*
 * per cpu array of lock stat structs
 */
typedef lstat_lock_counts_t	*lstat_cpu_counts_t;

typedef struct lstat_read_lock_counts {
	POINTER	    lock_ptr;            /* address of lock for output stats */
	uint32_t    read_lock_count;
	int64_t     cum_hold_ticks;       /* sum of read lock hold times over */
	                                  /* all callers. ....................*/
	uint32_t    write_index;          /* last write lock hash table index */
	uint32_t    busy_periods;         /* count of busy periods ended this */
	uint64_t    start_busy;           /* time this busy period started. ..*/
	uint64_t    busy_ticks;           /* sum of busy periods this lock. ..*/
	uint64_t    max_busy;             /* longest busy period for this lock*/
	uint32_t    max_readers;          /* maximum number of readers ...... */
#ifdef USER_MODE_TESTING
	rwlock_t    entry_lock;           /* lock for this read lock entry... */
	                                  /* avoid having more than one rdr at*/
	                                  /* needed for user space testing... */
	                                  /* not needed for kernel 'cause it  */
					  /* is non-preemptive. ............. */
#endif /* USER_MODE_TESTING */
} lstat_read_lock_counts_t;

/*
 * per cpu array of reader/writer lock stat structs
 */
typedef lstat_read_lock_counts_t	*lstat_read_lock_cpu_counts_t;

#endif /* CONFIG_LOCKMETER */

#endif /* _LINUX_LOCKMETER_H */
