/*
 *
 *  Copyright (C) 1999-2001 Silicon Graphics, Inc.
 *
 *  Written by John Hawkes (hawkes@sgi.com)
 *  Based on lockstat.c by Jack Steiner (steiner@sgi.com)
 *
 *  Modifications by Ray Bryant (raybry@us.ibm.com) Feb-Mar 2000
 *  Changes Copyright (C) 2000 IBM, Inc.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lockstat.h"

#define STRMAX		256
#define SPACEINC	16384*4

extern int  opt_debug;
char *strSpace(void);	/* allocate an efficient chunk for a string */

typedef struct {
	unsigned long	addr;
	char		*name;
} namespace_entry_t;

typedef struct {
	char              *tablename;
	namespace_entry_t *namespace;
	long               maxEntries;
	long               numEntries;
} namespace_t;

#define NAMESPACE_TABLE_CHUNK 2048

static namespace_t textSyms = {
	.tablename = "text",
	.namespace = NULL,
	.maxEntries = 0,
	.numEntries = 0
};

static namespace_t dataSyms = {
	.tablename = "data",
	.namespace = NULL,
	.maxEntries = 0,
	.numEntries = 0
};

/* for symbol -> address translation */
static namespace_t *allNamespaces[] = {
	&textSyms, &dataSyms, 0
};

/*
 * symbol mode/type from System.map or /proc/kallsyms
 */
#define TEXTMODE "tT"
#define DATAMODE "bBdDgGsD"	/* + "rRcC" ? */

//TODO:  if we include "read only" symbols [rR], need to filter out:
//       all syms with a '.' -- e.g., __func__.#
//       syms starting with:  __kstrtab_, __ksymtab_, __kcrctab_, ...
//       maybe syms starting with __param_ ?

/*
 * Open and read the entire map file once and build an in-memory table of the
 * address-name equivalences for quick reference.  If this is not the first
 * map file we've seen, then merge this file's address/symbol pairs with the
 * previous pairs.
 */
int
openMapFile (char *mapFilename)
{
#define MAX_MAPFILE_REC_LEN 128
	FILE *mapFile;
	char buffer[MAX_MAPFILE_REC_LEN];
	char func_name[MAX_MAPFILE_REC_LEN];
	char mode[MAX_MAPFILE_REC_LEN];
	unsigned long func_addr = 0;
	unsigned long max_addr = 0;
	int i, found_end = 0;

	namespace_t       *whichNamespace;
	namespace_entry_t *namespace;
	long               numNamespaceEntries;

	mapFile = fopen(mapFilename,"r");
	if (mapFile == NULL)
		return 0;	/* cannot open map file! */

#if 0	// not needed after fopen() ?
	rewind(mapFile);
#endif
	while (fgets(buffer, MAX_MAPFILE_REC_LEN, mapFile)) {
		/* Module map file begins with "Address" -- discard */
		if (strncmp(buffer, "Address", strlen("Address")) == 0)
			continue;

		if (sscanf(buffer, "%lx %s %s",
			   &func_addr, mode, func_name) != 3) {
		    die(1,"Corrupted mapfile\n");
		}
		if (max_addr < func_addr)
			max_addr = func_addr;

		if (strlen(mode) > 1) {
			/* Module map file has symbolic name as field #2 */
			strcpy(func_name, mode);
			whichNamespace = &textSyms;  /* use text namespace */
		} else {
			/* Assume System.map or /proc/kallsyms style map */
			if (strchr(TEXTMODE, *mode)) {
				whichNamespace = &textSyms;
			} else if (strchr(DATAMODE, *mode)) {
				whichNamespace = &dataSyms;
			} else
				continue;

		}

		/* Expand the in-memory namespace table, if necessary */
		namespace = whichNamespace->namespace;
		numNamespaceEntries = whichNamespace->numEntries;
		if (numNamespaceEntries == whichNamespace->maxEntries) {
		    whichNamespace->maxEntries += NAMESPACE_TABLE_CHUNK;
		    whichNamespace->namespace = namespace =
		        (namespace_entry_t *)realloc( (void *)namespace,
			(whichNamespace->maxEntries * sizeof(namespace_entry_t)));
		    if (!whichNamespace->namespace) {
			die(1, "Cannot malloc in-memory namespace table\n");
		    }
		}

		/*
		 *  Ensure monotomically increasing, non-duplicate addresses.
		 */
		if (numNamespaceEntries == 0) {
			/* the 1st entry */
			namespace[numNamespaceEntries].addr = func_addr;
			namespace[numNamespaceEntries].name = strSpace();
			strcpy(namespace[numNamespaceEntries].name, func_name);
			whichNamespace->numEntries = ++numNamespaceEntries;

		} else if (func_addr > namespace[numNamespaceEntries-1].addr) {
			/* this address is beyond the previous last address, so append */
			namespace[numNamespaceEntries].addr = func_addr;
			namespace[numNamespaceEntries].name = strSpace();
			strcpy(namespace[numNamespaceEntries].name, func_name);
			whichNamespace->numEntries = ++numNamespaceEntries;

#if 0 // not with separate text/data tables
			if (strcmp(func_name, "_end") == 0) {
				++found_end;
				break;
			}
#endif

		} else if (func_addr == namespace[numNamespaceEntries-1].addr) {
			/* duplicates the last entry; use this latter sym name */
#ifdef DEBUGSYM
			fprintf(stderr, "%s:  %s @ 0x%lx overwriting %s\n",
			       __FUNCTION__, func_name, func_addr,
			        namespace[numNamespaceEntries-1].name);
#endif
			strcpy(namespace[numNamespaceEntries-1].name, func_name);
		} else {
			/*
			 * This address preceeds the previous last address, so insert.
			 * Scan backwards from the end of the list to find the
			 * insertion point, moving entries forward one index position.
			 */
			whichNamespace->numEntries = ++numNamespaceEntries;
			for (i = numNamespaceEntries-2; i >= -1; i--) {
				if (i < 0  ||   func_addr > namespace[i].addr) {
					/* found the insertion point */
					namespace[i+1].addr = func_addr;		
					namespace[i+1].name = strSpace();
					strcpy(namespace[i+1].name, func_name);
					break;
				} else {
					namespace[i+1].addr = namespace[i].addr;
					namespace[i+1].name = namespace[i].name;
				}
			}
		}
	    }

#ifdef DEBUGSYM
	fprintf(stderr, "%s:  %ld text symbols, %ld data symbols\n",
	        __FUNCTION__, textSyms.numEntries, dataSyms.numEntries);
#endif
	fclose(mapFile);
	return 1;
}

/*
 * strSpace
 *
 * A crude allocation manager for string space. 
 * Returns pointer to where a string of length 0 .. STRMAX can be placed.
 * On next call, a free space pointer is updated to point to 
 * the location just beyond the last string allocated. 
 *	NOTE: "just beyond" could be 1 byte, but we increment to the
 *		next 8-byte boundary for cleaner copies and compares.
 *	NOTE: does not support expanding a string once allocated &
 *		another call to strspace is made.
 */
char*
strSpace(void)
{
	static char	*space=NULL, *endspace=NULL;
	int		incr;

	if (space) {
		/*
		 * The last caller might not have used this space yet, so
		 * assume the worst case: the max string.
		 */
		if (space[0] == 0) {
			incr = STRMAX;
		} else {
			incr = 9 + strlen(space);
			incr &= ~7;
		}
		space += incr;
	}

	if (space >= endspace) {
		space = (char *)malloc(SPACEINC);
		endspace = space + SPACEINC - STRMAX;
	}
	space[0] = '\0';
	return(space);
}

/*
 * translate an address (adr) into a name (if it matches) or to a name + offset
 * if it doesn't.  Return null if the name is below the lowest address in the
 * map file.  Return hex address if adr above '_end' or has "unreasonable" offset.
 * Need "resonable offset" filter for '-m /proc/kallsyms', because kallsyms has
 * no '_end' [on some archs, at least], but does include modules along with some
 * unreasonably large symbol values.
 *
 * Note:  data symbol table and text table loaded from /proc/kallsyms will not
 *        [should not?] have the '_end' entry.
 */
static unsigned long max_offset = 0x100000;
void
XlateAddressToSymbol(int datasym, void *adr, char *namep)
{
	int i;

	namespace_t       *whichNamespace = &textSyms;
	namespace_entry_t *namespace;

	if (datasym)
		whichNamespace = &dataSyms;

	if (opt_debug)
		namep += sprintf(namep, "[0x%lx] ", (unsigned long)adr);
	else
		namep[0] = '\0';/* ensure that subsequent strcat() will work */

	namespace = whichNamespace->namespace;

	for (i = 0; i < whichNamespace->numEntries; i++) {
	    if ((unsigned long)adr == namespace[i].addr) {
	       strcat(namep, namespace[i].name);	/* Bingo! */
	       return;
	    } else if ((unsigned long)adr < namespace[i].addr) {
		unsigned long offset;

		if (i == 0)	/* below the first address in the table */
		    break;

		if (strcmp(namespace[i-1].name, "_end") == 0)
		    break;	/* between "_end" and the modules -- just use numeric */

		offset = (unsigned long)adr - namespace[i-1].addr;
		if (offset > max_offset)
			break;	/* avoid unresonable offsets */

		sprintf(namep, "%s+0x%x", namespace[i-1].name, offset);
		return;
	    }
	}

	/* can't find it */
	if (!opt_debug)
		sprintf(namep, "[0x%lx]", (unsigned long)adr);
}
